# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


__author__ = "Norman Fomferra (Brockmann Consult GmbH)"

UNDEFINED_STR = "UNDEFINED"


class _Undefined:
    """
    Represents the UNDEFINED value.
    """
    _hash_code = hash(UNDEFINED_STR) + 1

    def __str__(self):
        return UNDEFINED_STR

    def __repr__(self):
        return UNDEFINED_STR

    def __eq__(self, other):
        return self is other or isinstance(other, _Undefined)

    def __hash__(self) -> int:
        return _Undefined._hash_code


#: Singleton value used to indicate an undefined state.
UNDEFINED = _Undefined()
